package com.icontrols.kakao.socket.wallpad;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import javax.annotation.Resource;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.icontrols.kakao.common.ERROR_INFO;
import com.icontrols.kakao.common.KAKAO_RESULT;
import com.icontrols.kakao.common.KEY_INFO;
import com.icontrols.kakao.common.THINGS_INFO;
import com.icontrols.kakao.common.TYPE;
import com.icontrols.kakao.common.utils.BeanUtils;
import com.icontrols.kakao.common.utils.CommonUtil;
import com.icontrols.kakao.common.utils.NodeUtil;
import com.icontrols.kakao.common.utils.XmlUtil;
import com.icontrols.kakao.config.AsyncConfig;
import com.icontrols.kakao.entity.ThingsInfo;
import com.icontrols.kakao.entity.UserInfo;
import com.icontrols.kakao.repo.ThingsInfoRepository;
import com.icontrols.kakao.socket.utils.IPGenerator;
import com.icontrols.kakao.socket.utils.SSLClient;
import com.icontrols.kakao.socket.utils.XMLConstants.REMOTE_KIND;
import com.icontrols.kakao.socket.utils.XMLGenerator;

@Service("kakaoMsgHandler")
public class KakaoMsgHandler {
	private final static Logger logger = LoggerFactory.getLogger(KakaoMsgHandler.class);
	
	@Resource(name="asyncConfig")
	private AsyncConfig asyncConfig;
	
	@Resource(name="wallpadMsgSender")
	WallpadMsgSender wallpadMsgSender;
	
	/**
	 * 카카오 전용 인증 프로세스
	 * 월패드에 인증번호를 노출시키는 메시지를 전송한다.
	 * @param param
	 * 		
	 * @return
	 */
	public Map<String, Object> doAuthProcess(Map<String, Object> param){
		Map<String, Object> retMap = new HashMap<String, Object>();
		
		try {
			int dong = Integer.parseInt((String)param.get("dong"));
			int ho = Integer.parseInt((String)param.get("ho"));
			String code = (String)param.get("code");
			
			String targetIP = IPGenerator.getIPFromDongHo(dong, ho);
			String twinid = "KAKAO"+dong+"_"+ho;
			String alias = "Kakao";
			StringBuffer sb = new StringBuffer();
			
			sb.append(XMLGenerator.MakeXMLHeaderForNotice(REMOTE_KIND.remote_register_new_device_auth_number.getContents()))
			.append("      <home_info dong=\"").append(dong).append("\" ho=\"").append(ho).append("\" ip=\"").append(targetIP).append("\" type=\"main\"/>\r\n")
			.append("      <device_info alias=\"").append(alias).append("\" twinid=\"").append(twinid).append("\" auth_number=\"" + code + "\" />\r\n")
			.append(XMLGenerator.MakeXMLTail());
			
			logger.info(sb.toString());
			
			String rcvMsg = SSLClient.sendToWallpad(targetIP, sb.toString());
			
			if(rcvMsg.trim().length() == 0) {
				retMap.put("result", ERROR_INFO.CODE.SUCCESS.getContents());
				retMap.put("resultCode", ERROR_INFO.CODE.SUCCESS.getmIdx());
			}
		}catch(Exception e) {
			retMap.put("result", ERROR_INFO.CODE.FAIL_999.getContents());
			retMap.put("resultCode", ERROR_INFO.CODE.FAIL_999.getmIdx());
			e.printStackTrace();
		}
		return retMap;
	}
	
	/**
	 * 기기 구독 / 해제 에 대한 메시지를 월패드에 전송한다. 
	 * 기기 구독이 돼 있지 않으면 이벤트를 받지 못한다.
	 * @param param
	 * @return
	 */
	public Map<String, Object> subscriptionProcess(Map<String, Object> param) {
		Map<String, Object> retMap = new HashMap<String, Object>();
		
		try {
			Document doc = new WallpadMsgSender().remoteDeviceSubscription(param);
			
			if(doc == null) {
				logger.error("[subscriptionProcess ERROR] - Document null");
				retMap.put(KEY_INFO.HTTP_KEY_ERROR, KAKAO_RESULT.getErrorInfo(KAKAO_RESULT.FAIL_UNREACHABLE));
				return retMap;
			}
			
			retMap.put("result", "success");
		}catch(Exception e) {
			logger.error("[subscriptionProcess ERROR] - " + e.getStackTrace());
			retMap.put(KEY_INFO.HTTP_KEY_ERROR, KAKAO_RESULT.getErrorInfo(KAKAO_RESULT.FAIL_INTERNAL));
		}
		return retMap;
	}
	
	public Map<String, Object> remoteRegisterNewDevice(Map<String, Object> param) { 
		Map<String, Object> retMap = new HashMap<String, Object>();
		
		try {
			int dong = Integer.parseInt((String)param.get("dong"));
			int ho = Integer.parseInt((String)param.get("ho"));
			String code = (String)param.get("code");
			
			String targetIP = IPGenerator.getIPFromDongHo(dong, ho);
			String twinid = "KAKAO_"+dong+"_"+ho;
			String alias = "KAKAO HOME";
			StringBuffer sb = new StringBuffer();
			
			sb.append(XMLGenerator.MakeXMLHeaderForSUBS("remote_register_new_device"))
			.append("      <home_info dong=\"").append(dong).append("\" ho=\"").append(ho).append("\" ip=\"").append(targetIP).append("\" type=\"main\"/>\r\n")
			.append("      <device_info alias=\"").append(alias).append("\" twinid=\"").append(twinid).append("\"  />\r\n")
			.append(XMLGenerator.MakeXMLTail());
			
			logger.info(sb.toString());
			
			SSLClient.sendToWallpad(targetIP, sb.toString());
			
			retMap.put("result", ERROR_INFO.CODE.SUCCESS.getContents());
			retMap.put("resultCode", ERROR_INFO.CODE.SUCCESS.getmIdx());
		}catch(Exception e) {
			retMap.put("result", ERROR_INFO.CODE.FAIL_999.getContents());
			retMap.put("resultCode", ERROR_INFO.CODE.FAIL_999.getmIdx());
			e.printStackTrace();
		}
		return retMap;
	}
	
	public List<ThingsInfo> getConnectDevices(int dong, int ho){
		List<ThingsInfo> thingList = new ArrayList<>();
		
		try {
			Document doc = new WallpadMsgSender().remoteDeviceStatusList(dong, ho);
			
			if(doc == null) {
				logger.error("[ERROR]-[Document is Null] - [" + dong + "-" + ho + "]");
				return null;
			}
			
			XPath xpath = XPathFactory.newInstance().newXPath();
			Node iMap = (Node)xpath.evaluate("//imap/service", doc, XPathConstants.NODE);
			String xmlResult = XmlUtil.getAttributeWithItem(iMap, "result");
			
			logger.debug("[KakaoMsgHandler - getConnectDevices] - xmlResult -> " + xmlResult);
			
			if(!xmlResult.equalsIgnoreCase("ok")) {
				logger.error("[KakaoMsgHandler - getConnectDevices] xmlResult not ok");
				return null;
			}
			
			// 월패드와 연결된 상위 THINGS 갯수 확인
			Node dCount =  null;
			NodeList devList = (NodeList)xpath.evaluate("//imap/service/devinfo", doc, XPathConstants.NODESET);
			
			NodeUtil.asList(devList)
					.stream()
					.filter(pNode -> THINGS_INFO.isAvailableThings(XmlUtil.getAttributeWithItem(pNode, "name")))
					.forEach(pNode -> {
						String tmpModelId = XmlUtil.getAttributeWithItem(pNode, "name");
						int childCount = Integer.parseInt(XmlUtil.getAttributeWithItem(pNode, "value"));
						
						logger.debug("tmpModelID --> " + tmpModelId);
						try {
							if(tmpModelId.equalsIgnoreCase("light") || tmpModelId.equalsIgnoreCase("electric")) {
								for(int j=1; j <= childCount; j++) {
									String childName = j < 10 ? tmpModelId+"0"+j : tmpModelId+j;
									Node childNode = (Node)xpath.evaluate("//imap/service/devinfo/*[@name='"+ childName +"']", doc, XPathConstants.NODE);
									logger.debug("childName = " + childName);
									int tmpChildCount = Integer.parseInt(XmlUtil.getAttributeWithItem(childNode, "value"));
									
									NodeList childList = (NodeList)xpath.evaluate("//imap/service/devinfo/subdevinfo/*[@name='"+ childName +"']", doc, XPathConstants.NODESET);
									
									thingList.addAll(getChildThingsInfoList(dong, ho, tmpChildCount, childList));
								}
							}else {
								NodeList childList = (NodeList)xpath.evaluate("//imap/service/devinfo/*[@name='"+ tmpModelId +"']", doc, XPathConstants.NODESET);
								thingList.addAll(getChildThingsInfoList(dong, ho, childCount, childList));
								
							}
						}catch(Exception e) {
							logger.error("[ERROR] at stream - " + e.getMessage());
							e.printStackTrace();
						}
					});
			thingList.add(new ThingsInfo("elevator","ElevatorCaller", Integer.toString(dong), Integer.toString(ho)));
		}catch(Exception e) {
			logger.error("[ERROR] - " + e.getMessage());
			return null;
		}
		
		return thingList;
	}
	/**
	 * 월패드와 연결된 기기들의 전체 상태를 월패드에 요청하고 응답 받는다.
	 * 지원 기기가 추가되면 해당 부분은 수정이 필요하다. 
	 * 자식 노드로 subdevinfo가 있는 기기들을 잘 확인해야 한다.
	 * @param dong
	 * @param ho
	 * @return
	 */
	
	public Map<String, Object> getDevicesStatusList(int dong, int ho) {
		Map<String, Object> retMap = new HashMap<String, Object>();
		Map<String, Object> body = new HashMap<String, Object>();
		List<String> unreachableIds = new ArrayList<String>();
		
		try {
			Document doc = new WallpadMsgSender().remoteDeviceStatusList(dong, ho);
			
			if(doc == null) {
				retMap.put(KEY_INFO.HTTP_KEY_ERROR, KAKAO_RESULT.getErrorInfo(KAKAO_RESULT.FAIL_UNREACHABLE));
				logger.error("[ERROR]-[Document is Null] - [" + dong + "-" + ho + "]");
				return retMap;
			}
			
			XPath xpath = XPathFactory.newInstance().newXPath();
			Node iMap = (Node)xpath.evaluate("//imap/service", doc, XPathConstants.NODE);
			String xmlResult = XmlUtil.getAttributeWithItem(iMap, "result");
			
			logger.debug("[KakaoMsgHandler - getDeviceList] - xmlResult -> " + xmlResult);
			
			if(!xmlResult.equalsIgnoreCase("ok")) {
				retMap.put(KEY_INFO.HTTP_KEY_ERROR, KAKAO_RESULT.getErrorInfo(KAKAO_RESULT.FAIL_UNREACHABLE));
				logger.error("[KakaoMsgHandler - getDeviceList] xmlResult not ok");
				return retMap;
			}
			
			// 월패드와 연결된 상위 THINGS 갯수 확인
			Node dCount =  null;
			NodeList devList = (NodeList)xpath.evaluate("//imap/service/devinfo", doc, XPathConstants.NODESET);
			
			NodeUtil.asList(devList)
					.stream()
					.filter(pNode -> THINGS_INFO.isAvailableThings(XmlUtil.getAttributeWithItem(pNode, "name")))
					.forEach(pNode -> {
						String tmpModelId = XmlUtil.getAttributeWithItem(pNode, "name");
						int childCount = Integer.parseInt(XmlUtil.getAttributeWithItem(pNode, "value"));
						
						List<Map<String, Object>> childNodeInfoList = null;
						logger.debug("tmpModelID --> " + tmpModelId + "childCount = " + childCount);
						try {
							// subdevinfo 를 가지고 있는 THINGS일 경우는 subdevinfo 들을 다시 구해야 한다.
							// 예를들어 light는 light01 ~ 05 등의 subdevinfo를 가진다.

							// livinglight와 light는 -> dimmer와 colorTemp를 조절이 가능한 조명이 포함될 수 있다.
							// 전등의 경우 월패드에서는 크게 livinglight 또는 light 로 나뉘고 속성은 같은반변 
							// 카카오에서는 light, LightDimmer, ColorLightDimmer 세가지로 나뉜다.
							// 월패드의 status는 (전원)/(디밍값)/(색상온도값) 으로 구성되며 
							// 전원을 제외하고 각 지원하지 않는 기능은 null로 전달된다. 
							// LightDimmer는 월패드에서 status가 on/40/null 로  status가 구성되고
							// ColorLightDimmer는 월패드에서 status가 on/40/10 의 형태 또는 status가 on/null/10 의형태로 구성된다. 
							// status의 구성요소에 맞게 capability가 구성되어야 한다. 
							
							
							if(tmpModelId.equalsIgnoreCase("light") || tmpModelId.equalsIgnoreCase("electric")) {
								for(int j=1; j <= childCount; j++) {
									String childName = j < 10 ? tmpModelId+"0"+j : tmpModelId+j;
									Node childNode = (Node)xpath.evaluate("//imap/service/devinfo/*[@name='"+ childName +"']", doc, XPathConstants.NODE);
									logger.debug("childName = " + childName + " // j = " + j);
									
									if(childNode == null) continue;
									
									String tmpValue = XmlUtil.getAttributeWithItem(childNode, "value");
									
									if(tmpValue == null) continue;
									
									int tmpchildCount = Integer.parseInt(tmpValue);
									
									NodeList childList = (NodeList)xpath.evaluate("//imap/service/devinfo/subdevinfo/*[@name='"+ childName +"']", doc, XPathConstants.NODESET);
									
									childNodeInfoList = getChildDeviceInfo(tmpchildCount, childList);
									logger.debug("child-->>" + childNodeInfoList.toString());
									if(childNodeInfoList != null)
										childNodeInfoList.forEach(tMap -> { body.putAll(tMap); });
								}
							}else {
								NodeList childList = (NodeList)xpath.evaluate("//imap/service/devinfo/*[@name='"+ tmpModelId +"']", doc, XPathConstants.NODESET);
								
								childNodeInfoList = getChildDeviceInfo(childCount, childList);
								
								logger.debug("child-->>" + childNodeInfoList.toString());
								
								if(childNodeInfoList != null)
									childNodeInfoList.forEach(tMap -> { body.putAll(tMap); });
							}
						}catch(Exception e) {
							retMap.put(KEY_INFO.HTTP_KEY_ERROR, KAKAO_RESULT.getErrorInfo(KAKAO_RESULT.FAIL_INTERNAL));
							logger.error("[ERROR] at stream - " + e.getMessage());
							e.printStackTrace();
						}
					});
			// elevator는 하드코딩.
			// elevator가 존재해야 카카오홈 앱에서 엘리베이터를 제어할 수 있다.
			Map<String, Object> eleMap = new HashMap<String, Object>();
			
			eleMap.put("displayType", THINGS_INFO.THINGS.ELEVATOR.getKakaoDisplayName());
			eleMap.putAll(getDeviceCapabilities(THINGS_INFO.THINGS.ELEVATOR.getThingsName(), null));
			eleMap.put("endpointId", THINGS_INFO.DEVICE_DETAIL.ELEVATOR.getModelId());
			eleMap.put("name", THINGS_INFO.THINGS.ELEVATOR.getDesc());
			eleMap.put("state", new HashMap<String, Object>());
			
			body.put("elevator", eleMap);
			
		}catch(Exception e) {
			retMap.put(KEY_INFO.HTTP_KEY_ERROR, KAKAO_RESULT.getErrorInfo(KAKAO_RESULT.FAIL_INTERNAL));
			logger.error("[ERROR] - " + e.getMessage());
		}
		
		retMap.put(KEY_INFO.HTTP_KEY_UNREACHABLEIDS, unreachableIds);
		retMap.put(KEY_INFO.HTTP_KEY_BODY, body);
		
		return retMap;
	}
	
	/**
	 * 동 호 정보 EndpointID 로 월패드에 상태 조회 메세지를 전송하고 결과를 반환 받는다. 
	 * @param dong
	 * @param ho
	 * @param endpointId
	 * @return
	 */
	public Map<String, Object> getDeviceStatus(int dong, int ho, String endpointId){
		Map<String, Object> retMap = new HashMap<String, Object>();
		String modelId = "";
		int devNum = 1;

		Document doc = null;

		// endpoint Id로 존재하는 THINGS 인지 확인 
		// 발생 ERROR FAIL_NONEXIST
		try {
			String[] tmpSplit = endpointId.split("_");
			
			modelId = tmpSplit[0];
			devNum = Integer.parseInt(tmpSplit[1]);	// 회로 번호
			
			if(!THINGS_INFO.isExistThings(modelId)) {
				retMap.put(KEY_INFO.HTTP_KEY_ERROR, KAKAO_RESULT.getErrorInfo(KAKAO_RESULT.FAIL_NONEXIST));
				logger.error("[ERROR]-[(" + endpointId + ") - Non Exist]");
				return retMap;
			}
		}catch(Exception e) { 
			retMap.put(KEY_INFO.HTTP_KEY_ERROR, KAKAO_RESULT.getErrorInfo(KAKAO_RESULT.FAIL_NONEXIST));
			logger.error("[ERROR]-[(" + endpointId + ") - Non Exist]");
			return retMap;
		}
		
		try {
			doc = new WallpadMsgSender().remoteDeviceStatus(dong, ho, modelId, devNum);
			
			if(doc == null) {
				retMap.put(KEY_INFO.HTTP_KEY_ERROR, KAKAO_RESULT.getErrorInfo(KAKAO_RESULT.FAIL_UNREACHABLE));
				logger.error("[ERROR]-[Document is Null] - [" + dong + "-" + ho + "] ["+ modelId +"]");
				return retMap;
			}
		}catch(Exception e) { 
			retMap.put(KEY_INFO.HTTP_KEY_ERROR, KAKAO_RESULT.getErrorInfo(KAKAO_RESULT.FAIL_NONEXIST));
			logger.error("[ERROR]-[(" + endpointId + ") - Non Exist]");
			return retMap;
		}
		
		return getStatusMap(endpointId, doc);
	}
	
	/**
	 * 월패드에 장치 제어 명령을 보내고 결과를 리턴한다.
	 * @param dong
	 * @param ho
	 * @param endpointId
	 * @param actionName
	 * @param desired
	 * @return
	 */
	public Map<String, Object> sendDeviceAction(int dong, int ho, String endpointId, String actionName, Map<String, Object> desired){
		Map<String, Object> retMap = new HashMap<String, Object>();
		
		String[] tmpSplit = endpointId.split("_");
		String modelId = (tmpSplit[0].equals("dimmer") || tmpSplit[0].equals("colorDimmer"))?tmpSplit[1]:tmpSplit[0];
		
		//엘리베이터 특정. -> 엘리베이터는 tmpSplit length가 1이다.
		// 모드 endpoint id는 _ <- 언더바가 포함돼 있지만 엘리베이터는 그렇지 않음
		// 엘리베이터가 아닌데도 tmpSplit length가 1인경우는 존재하지 않는 디바이스로 간주.
		if(tmpSplit.length == 1 && !modelId.equals(THINGS_INFO.DEVICE_DETAIL.ELEVATOR.getModelId())) {
			logger.error("[sendDeviceAction ERROR] - this is NonExist => " + modelId);
			retMap.put(KEY_INFO.HTTP_KEY_ERROR, KAKAO_RESULT.getErrorInfo(KAKAO_RESULT.FAIL_NONEXIST));
			return retMap;
		}
		
		int devNum = 1;
		
		if(tmpSplit[0].equals("dimmer") || tmpSplit[0].equals("colorDimmer")) devNum = Integer.parseInt(tmpSplit[2]);
		else if(modelId.equals(THINGS_INFO.DEVICE_DETAIL.ELEVATOR.getModelId())) devNum = 1;
		else devNum = Integer.parseInt(tmpSplit[1]);
		THINGS_INFO.THINGS things = THINGS_INFO.getThingsType(endpointId);
		
		if(things == null) {
			retMap.put(KEY_INFO.HTTP_KEY_ERROR, KAKAO_RESULT.getErrorInfo(KAKAO_RESULT.FAIL_INVALID_REQ));
			logger.error("[ERROR] - things null ==> INVALID REQ");
			return retMap;
		}
		
		String action = KakaoMsgHandler.setDeviceActionStr(endpointId, modelId, actionName, desired);
		
		if(action.length() == 0) {
			retMap.put(KEY_INFO.HTTP_KEY_ERROR, KAKAO_RESULT.getErrorInfo(KAKAO_RESULT.FAIL_INVALID_REQ));
			logger.error("[ERROR] - actions length --> 0 IVALID REQ " );
			return retMap;
		}
		
		Document doc = new WallpadMsgSender().remoteDeviceControl(
																dong, 
																ho, 
																modelId, 
																devNum, 
																action
															);
		
		if(doc == null) {
			retMap.put(KEY_INFO.HTTP_KEY_ERROR, KAKAO_RESULT.getErrorInfo(KAKAO_RESULT.FAIL_UNREACHABLE));
			return retMap;
		}
		
		return getStatusMap(endpointId, doc);
	}
	
	/**
	 * 상태조회 요청 메세지에 대한 응답 메세지를 파싱하여 카카오 형식으로 변환한다.
	 * @param endpointId
	 * @param doc
	 * @return
	 */
	public static Map<String, Object> getStatusMap(String endpointId, Document doc){
		Map<String, Object> retMap = new HashMap<String, Object>();
		
		XPath xpath = XPathFactory.newInstance().newXPath();
		try {
			Node iMap = (Node) xpath.evaluate("//imap/service", doc, XPathConstants.NODE);
			String xmlResult = XmlUtil.getAttributeWithItem(iMap, "result");
			
			if(!xmlResult.equalsIgnoreCase("ok")) {
				retMap.put(KEY_INFO.HTTP_KEY_ERROR, KAKAO_RESULT.getErrorInfo(KAKAO_RESULT.FAIL_INVALID_REQ)); 
				logger.error("[ERROR]-[xmlResult is not ok");
				return retMap;
			}
			
			String status = (String) xpath.evaluate("//imap/service/status", doc, XPathConstants.STRING);
			
			Map<String, Object> state = setDeviceState(endpointId, status);
			retMap.put(endpointId, state);
			
		} catch (XPathExpressionException e) {
			retMap.put(KEY_INFO.HTTP_KEY_ERROR, KAKAO_RESULT.getErrorInfo(KAKAO_RESULT.FAIL_INTERNAL));
			e.printStackTrace();
		}
		
		logger.debug("[UPDATE RESULT] - " + retMap.toString());
		return retMap;
	}
	
	
	/**
	 * 이벤트 발생시 카카오에 전송하는 Body를 반환한다.
	 * @param dong
	 * @param ho
	 * @param body
	 * @return
	 * @throws Exception
	 */
	public Map<String, Object> setKakaoEventMsg(UserInfo userInfo, Map<String, Object> body) throws Exception{
		Map<String, Object> retMap = new HashMap<String, Object>();
		List<String> unreachableIds = new ArrayList<>();
		
		retMap.put(KEY_INFO.HTTP_KEY_ID, UUID.randomUUID().toString()); 
		retMap.put(KEY_INFO.HTTP_KEY_TYPE, TYPE.REQ_TYPE.REPORTSTATE.getTypeName());
		retMap.put(KEY_INFO.HTTP_KEY_AUTHORIZATION, userInfo.getKakaoAccessToken());
		retMap.put(KEY_INFO.HTTP_KEY_VERSION, "1");
		retMap.put(KEY_INFO.HTTP_KEY_UNREACHABLEIDS, unreachableIds);
		retMap.put(KEY_INFO.HTTP_KEY_BODY, body);
		
		return retMap;
	}
	
	public Map<String, Object> setKDeleteIdsMsg(String dong, String ho, UserInfo userInfo){
		ThingsInfoRepository tRepo = (ThingsInfoRepository)BeanUtils.getBean("thingsInfoRepository");
		
	
		List<ThingsInfo> tList = tRepo.findByDongAndHo(dong, ho);
		
		List<String> endpointIds = new ArrayList<String>();
		
		for(ThingsInfo tInfo:tList) endpointIds.add(tInfo.getEndpointId());
		
		Map<String, Object> retMap = new HashMap<String, Object>();
		
		retMap.put(KEY_INFO.HTTP_KEY_ID, UUID.randomUUID().toString()); 
		retMap.put(KEY_INFO.HTTP_KEY_TYPE, TYPE.REQ_TYPE.REPORTSTATE.getTypeName());
		retMap.put(KEY_INFO.HTTP_KEY_AUTHORIZATION, userInfo.getKakaoAccessToken());
		retMap.put(KEY_INFO.HTTP_KEY_VERSION, "1");
		retMap.put(KEY_INFO.HTTP_KEY_DELETEDIDS, endpointIds);
		
		return retMap;
	}
	/**
	 * 지원하는 기기들의 Capability 를 카카오 형식에 맞게 추출한다.
	 * 기기가 추가되게 되면 목록을 수정해야한다.
	 * @param modelId
	 * @return
	 */
	public static Map<String, Object> getDeviceCapabilities(String modelId, String status){
		THINGS_INFO.THINGS things = THINGS_INFO.getThingsType(modelId);
		
		String jsonString = "";
		
		switch(things.getEndPoint()) {
		case THINGS_INFO.THINGS_LIGHT:
			jsonString = THINGS_INFO.THINGS_CAPABILITY.LIGHT.getCapJsonStr();
			break;
		case THINGS_INFO.THINGS_LIGHT_DIMMER:
			jsonString = THINGS_INFO.THINGS_CAPABILITY.LIGHT_DIMMER.getCapJsonStr();
			break;
			
		case THINGS_INFO.THINGS_COLOR_LIGHT_DIMMER:
			String[] splitStr = status.split("/");
			if(splitStr[1].equals("null"))
				jsonString = THINGS_INFO.THINGS_CAPABILITY.COLOR_LIGHT.getCapJsonStr();
			else
				jsonString = THINGS_INFO.THINGS_CAPABILITY.COLOR_LIGHT_DIMMER.getCapJsonStr();
			break;
			
		case THINGS_INFO.THINGS_TEMPER:
			jsonString = THINGS_INFO.THINGS_CAPABILITY.TEMPER.getCapJsonStr();
			break;
			
		case THINGS_INFO.THINGS_GAS:
			jsonString = THINGS_INFO.THINGS_CAPABILITY.GAS.getCapJsonStr();
			break;
			
		case THINGS_INFO.THINGS_VENTIL:
			jsonString = THINGS_INFO.THINGS_CAPABILITY.VENTIL.getCapJsonStr();
			break;
			
		case THINGS_INFO.THINGS_ELEVATOR:
			jsonString = THINGS_INFO.THINGS_CAPABILITY.ELEVATOR.getCapJsonStr();
			break;
			
		case THINGS_INFO.THINGS_ELECPLUG:
			jsonString = THINGS_INFO.THINGS_CAPABILITY.ELECTRIC.getCapJsonStr();
			break;
		}
		Map<String, Object> capabilities = CommonUtil.jsonToMap(jsonString);
		
		return capabilities;
	}
	
	public List<ThingsInfo> getChildThingsInfoList(int dong, int ho, int childCount, NodeList childList) {
		List<ThingsInfo> thingsList = new ArrayList<>();
		
		try {
			NodeUtil.asList(childList)
			.stream()
			.filter(cNode -> cNode != null)
			.forEach(cNode -> {
				try {
					Map<String, Object> deviceMap = new HashMap<>();
					String cNodeName = XmlUtil.getAttributeWithItem(cNode, "name");
					String cNodeDevNum = XmlUtil.getAttributeWithItem(cNode, "dev_num");
					String status = XmlUtil.getElemetTextValue(cNode);
					String endpointId = cNodeName+"_"+cNodeDevNum;
					
					String tmpModelId = endpointId;
					
					if(cNodeName.contains("light")) {
						String[] strSplit = status.split("/");
						// split 이 안됐다면 on off만 되는 예전 방식으로 전등 상태를 받았다는 뜻.
						if(strSplit.length > 1) {
							if(!strSplit[1].equals("null") && strSplit[2].equals("null")) {
								endpointId = "dimmer_"+endpointId;
							}
							else if(!strSplit[2].equals("null")) {
								endpointId = "colorDimmer_"+endpointId;
							}
							tmpModelId = cNodeName+"_"+cNodeDevNum;
						}
					}
					
		 			
					logger.debug("tmpModelId = " + tmpModelId + " || endpointID = " + endpointId);
					
		    		ThingsInfo thingsInfo = new ThingsInfo(endpointId, tmpModelId, Integer.toString(dong), Integer.toString(ho));
		    		
		    		thingsList.add(thingsInfo);
		    		
				}catch(Exception e) {
					e.printStackTrace();
				}
			});	
		}catch(Exception e) { 
			return null;
		}
		
		return thingsList;
	}
	
	/**
	 * Document에서 추출한 Device의 자식 노드들을 추출한다.
	 * @param childCount
	 * @param childList
	 * @return
	 */
	public List<Map<String, Object>> getChildDeviceInfo(int childCount, NodeList childList){
		List<Map<String, Object>> deviceList = new ArrayList<>();
		try {
			NodeUtil.asList(childList)
				.stream()
				.filter(cNode -> cNode != null)
				.forEach(cNode -> {
					try {
						Map<String, Object> deviceMap = new HashMap<>();
						String cNodeName = XmlUtil.getAttributeWithItem(cNode, "name");
						String cNodeDevNum = XmlUtil.getAttributeWithItem(cNode, "dev_num");
						String status = XmlUtil.getElemetTextValue(cNode);
						String endpointId = cNodeName+"_"+cNodeDevNum;
						
						if(cNodeName.contains("light")) {
							String[] strSplit = status.split("/");
							// split 이 안됐다면 on off만 되는 예전 방식으로 전등 상태를 받았다는 뜻.
							if(strSplit.length > 1) {
								if(!strSplit[1].equals("null") && strSplit[2].equals("null")) {
									endpointId = "dimmer_"+endpointId;
									cNodeName = "dimmer_" + cNodeName;
								}
								else if(!strSplit[2].equals("null")) {
									endpointId = "colorDimmer_"+endpointId;
									cNodeName = "colorDimmer_"+cNodeName;
								}
							}
						}
						
			 			String tmpModelId = cNodeName.equals("temper")?endpointId:cNodeName;
						logger.debug("tmpModelId = " + tmpModelId + " || endpointID = " + endpointId);
			 			
						
						THINGS_INFO.DEVICE_DETAIL detailDev = THINGS_INFO.getDeviceInfo(tmpModelId, status);
						
						Map<String, Object> deviceInfo =  new HashMap<String, Object>();
						
						deviceInfo.put(KEY_INFO.BODY_KEY_ENDPOINT_ID, endpointId);
			    		deviceInfo.put(KEY_INFO.BODY_KEY_NAME, detailDev.getDisplayName()+cNodeDevNum);
			    		deviceInfo.put(KEY_INFO.BODY_KEY_DISPLAY_TYPE, detailDev.getDisplayType());
			    		
			    		
			    		deviceInfo.putAll(getDeviceCapabilities(endpointId, status));
			    		deviceInfo.put(KEY_INFO.BODY_KEY_STATE, setDeviceState(endpointId, status));
			    		
			    		deviceMap.put(endpointId, deviceInfo);
			    		deviceList.add(deviceMap);
			    		
					}catch(Exception e) {
						e.printStackTrace();
					}
					
				});			
		}catch(Exception e) {
			return null;
		}
		return deviceList;
	}
	
	public List<Map<String, Object>> getChildDeviceInfo_bak(int childCount, NodeList childList){
		List<Map<String, Object>> deviceList = new ArrayList<>();
		
		for(int j=0; j < childCount; j++) {
			Map<String, Object> deviceMap = new HashMap<>();
			
			Node cNode = childList.item(j);
			if(cNode == null) {
				logger.error("cNode is null");
				return null;
			}
			
			String cNodeName = XmlUtil.getAttributeWithItem(cNode, "name");
			String cNodeDevNum = XmlUtil.getAttributeWithItem(cNode, "dev_num");
			String status = XmlUtil.getElemetTextValue(cNode);
			String endpointId = cNodeName+"_"+cNodeDevNum;
			
 			String tmpModelId = cNodeName.equals("temper")?endpointId:cNodeName;
			
			THINGS_INFO.DEVICE_DETAIL detailDev = THINGS_INFO.getDeviceInfo(tmpModelId, status);
			
			Map<String, Object> deviceInfo =  new HashMap<String, Object>();
			
			THINGS_INFO.THINGS things = THINGS_INFO.getThingsType(cNodeName);
			
			if(things.getEndPoint() == THINGS_INFO.THINGS_COLOR_LIGHT_DIMMER)
				endpointId = "colorDimming_"+endpointId;
			
			if(things.getEndPoint() == THINGS_INFO.THINGS_LIGHT_DIMMER) 
				endpointId = "dimming_"+endpointId;
			
			
			deviceInfo.put(KEY_INFO.BODY_KEY_ENDPOINT_ID, endpointId);
    		deviceInfo.put(KEY_INFO.BODY_KEY_NAME, detailDev.getDisplayName()+cNodeDevNum);
    		deviceInfo.put(KEY_INFO.BODY_KEY_DISPLAY_TYPE, detailDev.getDisplayType());
    		
    		deviceInfo.putAll(getDeviceCapabilities(endpointId, status));
    		deviceInfo.put(KEY_INFO.BODY_KEY_STATE, setDeviceState(endpointId, status));
    		
    		deviceMap.put(endpointId, deviceInfo);
    		deviceList.add(deviceMap);
		}
		
		return deviceList;
	}
	
	/**
	 * 디바이스의 상태 명령어를 디바이스 종류에 맞게 생성한다.
	 * 카카오에서 던져준 capability 값부터 유효성 검사를 한 뒤에 해당 capability와 맞는 속성인지도 확인해야 한다.
	 * @param modelId
	 * @param desired
	 * @return
	 */
	
	public static String setDeviceActionStr(String endPointId, String modelId, String capability, Map<String, Object> desired) {
		StringBuffer sb = new StringBuffer();
		
		THINGS_INFO.THINGS things = THINGS_INFO.getThingsType(endPointId);
		
		logger.debug("[setDeiceActionStr] - modelId = " + modelId + " || capability = " + capability + " || things endpoint = " + things.getEndPoint());
		logger.debug("ACTION STATUS --> " + desired.toString());
		try {
			int brightness = 1;
			int colorTemper = 1;
			switch(things.getEndPoint()) {
				
			case THINGS_INFO.THINGS_LIGHT:
				if(capability.equals("Power") && desired.containsKey("power")) {
					String tmpState = (String)desired.get("power");
					if(tmpState.equals("on") || tmpState.equals("off")) sb.append(tmpState);
					
					sb.append("/null/null");
				}
				
				break;
			case THINGS_INFO.THINGS_LIGHT_DIMMER:
				
				if(capability.equals("Power") && desired.containsKey("power")) {
					String tmpState = (String)desired.get("power");
					if(tmpState.equals("on") || tmpState.equals("off")) {
						sb.append(tmpState);
						sb.append("/null/null");
					}
				}
				
				if(desired.containsKey("brightness")) {
					brightness = ((int)desired.get("brightness"))/10 == 0 ? 1: ((int)desired.get("brightness"))/10;
					sb.append("on").append("/").append(brightness).append("/null");
				}
				
				break;
				
			case THINGS_INFO.THINGS_COLOR_LIGHT_DIMMER:
				
				if(capability.equals("Power") && desired.containsKey("power")) {
					String tmpState = (String)desired.get("power");
					if(tmpState.equals("on") || tmpState.equals("off")) {
						sb.append(tmpState);
					}
				}
				
				
				if(desired.containsKey("brightness")) {
					brightness = ((int)desired.get("brightness"))/10 == 0 ? 1: ((int)desired.get("brightness"))/10;
					if(desired.containsKey("power"))
						sb.append("/").append(brightness);
					else
						sb.append("on/").append(brightness);
				}else {
					sb.append("/null");
				}
				
				
				if(desired.containsKey("colorTemperature")) {
//					colorTemper = ((int)desired.get("colorTemperature"))/1000 == 0 ? 1:((int)desired.get("colorTemperature"))/1000;
					int tmpColorTemp = (int)desired.get("colorTemperature");
					
					if(tmpColorTemp < 3000) tmpColorTemp = 3000;
					else if(tmpColorTemp > 5700) tmpColorTemp = 5700;
					
					colorTemper = (tmpColorTemp)/300 - 9;
					if(desired.containsKey("power"))
						sb.append("/").append(colorTemper);
					else {
						if(desired.containsKey("brightness")) {
							sb.append("/").append(colorTemper);	
						}else {
							sb = new StringBuffer();
							sb.append("on/null/").append(colorTemper);
						}
						
					}
				}else {
					sb.append("/null");
				}
				
				
				break;
				
			case THINGS_INFO.THINGS_TEMPER:
				if(capability.equals("Power") && desired.containsKey("power")) {
					String tmpState = (String)desired.get("power");
					if(tmpState.equals("on") || tmpState.equals("off")) sb.append(tmpState);
					
					if(tmpState.equals("on") && desired.containsKey("targetSetpoint")) sb.append("/").append(desired.get("targetSetpoint"));
				}
					
				if(capability.equals("TemperatureSetpoint") && desired.containsKey("targetSetpoint")) {
					sb.append("on/").append(desired.get("targetSetpoint"));
				}
				
				
				break;
				
			case THINGS_INFO.THINGS_GAS:
				if(capability.equals("Lock") && desired.containsKey("locked")){
					if((boolean)desired.get("locked"))
						sb.append("close");
				}
				break;
				
			case THINGS_INFO.THINGS_VENTIL:
				if(capability.equals("Power") && desired.containsKey("power")) {
					String tmpState = (String)desired.get("power");
					if(tmpState.equals("on") || tmpState.equals("off")) sb.append(tmpState);
					
					if(tmpState.equals("on") && desired.containsKey("fanSpeed")) {
						String tmpFanSpeed = (String)desired.get("fanSpeed");
						if(tmpFanSpeed.equals("low") || tmpFanSpeed.equals("middle") || tmpFanSpeed.equals("high"))
							sb.append(desired.get("fanSpeed"));
					}
				}
				
				if(capability.equals("AirFlow") && desired.containsKey("fanSpeed")) {
					String tmpFanSpeed = (String)desired.get("fanSpeed");
					
					if(tmpFanSpeed.equals("low") || tmpFanSpeed.equals("middle") || tmpFanSpeed.equals("high"))
						sb.append("on/").append(desired.get("fanSpeed"));
				}
				break;
				
			case THINGS_INFO.THINGS_ELECPLUG: // 지원하지 않음
				if(capability.equals("Power") && desired.containsKey("power")) {
					String tmpState = (String)desired.get("power");
					if(tmpState.equals("on"))
						sb.append("on");
					else if(tmpState.equals("off")) sb.append("off");
				}
				break;
				
			case THINGS_INFO.THINGS_ELEVATOR:
				if(capability.equals("ElevatorCall") && desired.containsKey("callDirection")) {
					if("down".equalsIgnoreCase((String)desired.get("callDirection"))) {
						sb.append(desired.get("callDirection"));
					}
				}
			}
		}catch(Exception e) {
			
		}
		
		return sb.toString();
	}
	
	/**
	 * 상태조회 이후 월패드에서 받은 상태 값을 디바이스의 속성에 맞게 파싱하고 
	 * Map에 <"속성명", "상태값"> 형태로 담아준다. 
	 * @param modelId
	 * @param status
	 * @return
	 */
	public static Map<String, Object> setDeviceState(String modelId, String status){
		Map<String,Object> state = new HashMap<String, Object>();
		
		THINGS_INFO.THINGS things = THINGS_INFO.getThingsType(modelId);
		String[] splitStr = status.split("\\/");
		int temp1, temp2 = 0;
		switch(things.getEndPoint()) {
		case THINGS_INFO.THINGS_ELECPLUG:
			state.put("power", splitStr[1]);
			
			break;
		case THINGS_INFO.THINGS_LIGHT:
		case THINGS_INFO.THINGS_LIGHT_DIMMER:
		case THINGS_INFO.THINGS_COLOR_LIGHT_DIMMER:
			
			state.put("power", splitStr[0]);
			
//			if(things.getEndPoint() == THINGS_INFO.THINGS_LIGHT) break; 
			if(splitStr.length == 1) break;
			
			if(splitStr[1] != null && !splitStr[1].equals("null")) {
				temp1 = Integer.parseInt(splitStr[1]) * 10;
				state.put("brightness", temp1);
			}
			
			if(splitStr[2] != null && !splitStr[2].equals("null")) {
				temp2 = ((Integer.parseInt(splitStr[2])+9)*300);
				state.put("colorMode", "colorTemperature");
				state.put("colorTemperature", temp2);
			}
			break;
		case THINGS_INFO.THINGS_TEMPER:
			
			state.put("power", splitStr[0]);
			state.put("targetSetpoint", Double.parseDouble(splitStr[1]));
			state.put("currentTemperature", Double.parseDouble(splitStr[2]));
			
			break;
		case THINGS_INFO.THINGS_GAS:
			if(status.equalsIgnoreCase("close"))
				state.put("locked", true);
			else
				state.put("locked", false);
			break;
			
		case THINGS_INFO.THINGS_VENTIL:
			state.put("power", splitStr[0]);
			if(splitStr[0].equalsIgnoreCase("on")) {
				if(splitStr[1].equals("mid")) state.put("fanSpeed", "middle");
				else state.put("fanSpeed", splitStr[1]);
			}
			break;
		case THINGS_INFO.THINGS_ELEVATOR:
			state.put("callDirection", "down");
			break;
		}
		
		return state;
	}
}
